; Bierzähler 1.0
;
; Copyright (C) 2009 Paul Wilhelm <paul@mosfetkiller.de>
; http://mosfetkiller.de/?s=bierzaehler
;
; Modul: main.asm | Hauptprogramm

.include "tn13def.inc"	; ATtiny13-spezifische Definitionen

.equ BUTTON = PB4		; Taster
.equ CLOCK = PB2		; CLOCK (Schieberegister)
.equ DATA = PB1			; DATA (Schieberegister)
.equ STROBE = PB0		; STROBE (Schieberegister)

.def DIGIT_0 = r20		; Einerstelle
.def DIGIT_1 = r21		; Zehnerstelle
.def DIGIT_2 = r22		; Hunderterstelle

.def MUX_DIGIT = r23	; Aktuell zu multiplexende Ziffer: 0-2
.def MUX_SEGMENT = r24	; Aktuell zu multiplexendes Segment der aktiven Ziffer: 2 ^ [0-6] (Bitmaske!)

.def COUNTER = r25		; Zählerstand

.def BUTTON_COUNTER = r26	; Tastenzähler

.def BUTTON_HOLD = r27	; Halteflag
.def HOLD_TIME = r28	; Haltezeit

.def POWER_OFF_PHASE = r29		; Power-Off Phase

; Uptimezähler
.def IDLE_QSECONDS = r1	; Viertelsekundenzähler
.def IDLE_MINUTES = r2	; Minutenzähler

.equ POWER_OFF_HOLD_TIME_1 = 8	; Power-Off Phase 1 - Zeitspanne: 8 * 250 ms = 2 s
.equ POWER_OFF_HOLD_TIME_2 = 12	; Power-Off Phase 2 - Zeitspanne: 12 * 250 ms = 3 s

.equ IDLE_AUTOSHUTDOWN = 2	; Nach ca. 2 Minuten (wirklich sehr sehr ungenau...sind eher 1:20 min) automatisch abschalten


; Interruptvektor
.cseg
.org 0
		rjmp   RESET			; Reset Handler
	reti;rjmp   EXT_INT0_vect	; IRQ0 Handler
	reti;rjmp   PCINT0_vect		; PCINT0 Handler
	reti;rjmp   TIM0_OVF_vect	; Timer0 Overflow Handler
	reti;rjmp   EE_RDY_vect		; EEPROM Ready Handler
	reti;rjmp   ANA_COMP_vect	; Analog Comparator Handler
		rjmp   TIM0_COMPA_vect	; Timer0 CompareA Handler
	reti;rjmp   TIM0_COMPB_vect	; Timer0 CompareB Handler
	reti;rjmp   WATCHDOG_vect	; Watchdog Interrupt Handler
	reti;rjmp   ADC_vect		; ADC Conversion Handler


; PCINT0 Handler (Aufwachpunkt nach dem Power-down Mode)
PCINT0_vect:
	reti


; Timer0 CompareA Handler (Uptime- und Haltezeitzähler)
TIM0_COMPA_vect:
	; Flags und in diesem Abschnitt verwendete Register sichern - ganz wichtig!
	push	r17
	push	r16
	in		r16, SREG

	; Viertelsekundenzähler erhöhen (kein Bereichscheck!)
	inc		IDLE_QSECONDS

	; Ist eine volle Minute erreicht?
	ldi		r17, 240	; 240 Viertelsekunden = 1 Minute
	cp		IDLE_QSECONDS, r17
	brlo	NO_NEW_MINUTE	; Wenn nicht, weiter zum Haltezeitzähler

	; Minutenzähler erhöhen (kein Bereichscheck!)
	inc		IDLE_MINUTES
	clr		IDLE_QSECONDS


NO_NEW_MINUTE:
	; Wird der Taster gehalten?
	cpi		BUTTON_HOLD, 0
	breq	TIM0_COMPA_vect_RETURN

	; Haltezeitzähler erhöhen, wenn HOLD_TIME < 255
	cpi		HOLD_TIME, 255
	breq	TIM0_COMPA_vect_RETURN

	inc		HOLD_TIME


TIM0_COMPA_vect_RETURN:
	; Flags wiederherstellen
	out		SREG, r16
	pop		r16
	pop		r17

	reti


; Initialisierungen
RESET:
	; Stack
	ldi		r16, low(RAMEND)
	out		SPL, r16

	; PORTB
	ldi		r16, (1 << CLOCK) | (1 << DATA) | (1 << STROBE)
	out		DDRB, r16

	ldi		r16, (1 << BUTTON)	; Taster-Pull-Up an; DATA = CLOCK = STROBE = L
	out		PORTB, r16

	; Display- und Multiplexing
	ldi		DIGIT_0, 0
	ldi		DIGIT_1, 0
	ldi		DIGIT_2, 0

	ldi		MUX_DIGIT, 0
	ldi		MUX_SEGMENT, (1 << (0))

	ldi		COUNTER, 0

	ldi		BUTTON_COUNTER, 0

	; Pin Change Interrupt
	ldi		r16, (1 << PCIE)	; Enable
	out		GIMSK, r16

	ldi		r16, (1 << BUTTON)	; PCINT für Taster aktivieren
	out		PCMSK, r16

	; Timer0 CompareA Handler
	ldi		r16, (1 << WGM01)	; CTC
	out		TCCR0A, r16

	ldi		r16, (1 << CS02) | (1 << CS00)	; CK/1024
	out		TCCR0B, r16

	ldi		r16, (1 << OCIE0A)	; Enable
	out		TIMSK0, r16

	ldi		r16, 244			; Vergleichswert (Periodendauer ca. 250 ms)
	out		OCR0A, r16

	sei


; Hauptprogramm
MAIN:
	; Evtl. in Leerzeichen umgewandelte (siehe weiter unten vor dem
	; Multiplexing), führende Nullen wieder in Nullen zurückverwandeln
	cpi		DIGIT_1, 10
	brne	NO_HIDDEN_ZERO
	ldi		DIGIT_1, 0


NO_HIDDEN_ZERO:
	; Prüfen, ob der Button zur Zeit gedrückt ist
	sbic	PINB, BUTTON
	rjmp	NO_BUTTON_PRESS


; Button ist gedrückt; Tastenzähler erhöhen (obere Grenze 255)
BUTTON_PRESS:
	cpi		BUTTON_COUNTER, 255
	breq	BUTTON_DONE

	inc		BUTTON_COUNTER
	rjmp	BUTTON_DONE

; Button ist nicht gedrückt; Tastenzähler erniedrigen (untere Grenze 0)
NO_BUTTON_PRESS:
	cpi		BUTTON_COUNTER, 0
	breq	BUTTON_DONE

	dec		BUTTON_COUNTER


BUTTON_DONE:
	; Ist dies der erste Druck auf den Taster oder wird die Taste noch gehalten?
	cpi		BUTTON_HOLD, 0
	breq	FIRST_PRESS


; Taste wird noch gehalten
HOLD:
	cp		BUTTON_COUNTER, BUTTON_HOLD
	brlo	CLEAR_HOLD

	; Haltezeit auf Zeitspanne für Power-Off Phase 1 überprüfen
	cpi		HOLD_TIME, POWER_OFF_HOLD_TIME_1
	brlo	MUX	; Ab zum Multiplexing, wenn Zeitspanne noch nicht erreicht

	; Haltezeit auf Zeitspanne für Power-Off Phase 2 überprüfen
	cpi		HOLD_TIME, POWER_OFF_HOLD_TIME_2
	brlo	POWER_OFF_1 ; Power-Off Phase 1, wenn Zeitspanne noch nicht erreicht


	; Power-Off Phase 2
	; In Schlafmodus schalten, nur von einem Tasterdruck wieder wecken lassen
	POWER_OFF_2:
		; Phase nicht erneut ausführen, wenn bereits getan
		cpi		POWER_OFF_PHASE, 2
		breq	MUX

		ldi		POWER_OFF_PHASE, 2

		; Schieberegister mit Nullbits füllen (Display ausschalten)
		rcall	SHIFTREG_CLEAR

		; Warten, bis der Taster wieder losgelassen wurde und Taster entprellen
		rcall	BUTTON_DEBOUNCE_WAIT

		; Abschalten (Schlafmodus)
		rcall	SHUTDOWN

		; Und hier geht es nach dem Aufwachen dann auch schon gleich wieder weiter

		; Warten, bis der Taster wieder losgelassen wurde und Taster entprellen
		rcall	BUTTON_DEBOUNCE_WAIT

		; Idle-Zeit resetten
		clr		IDLE_QSECONDS
		clr		IDLE_MINUTES

		rjmp	MUX


	; Power-Off Phase 1
	; Der vor dem Ausschalten ungewünscht erhöhte Zählerstand wird wieder verringert
	POWER_OFF_1:
		; Phase nicht erneut ausführen, wenn bereits getan
		cpi		POWER_OFF_PHASE, 1
		breq	MUX

		ldi		POWER_OFF_PHASE, 1

		rcall	DEC_COUNTER

		rjmp	MUX


; Der Taster wurde losgelassen
CLEAR_HOLD:
	ldi		BUTTON_HOLD, 0
	ldi		HOLD_TIME, 0
	ldi		BUTTON_COUNTER, 0
	ldi		POWER_OFF_PHASE, 0

	rjmp	MUX


; Die steigende Flanke des gehaltenen Tasters erkennen
FIRST_PRESS:
	; Hat der Tastenzähler bereits einen annehmbaren Wert erreicht?
	cpi		BUTTON_COUNTER, 200	; Schwellwert für registrierten Tastendruck
	brlo	MUX	; Wenn nicht, ab zum Multiplexing

	; Ansonsten das Halteflag setzen
	ldi		BUTTON_HOLD, 180	; 200 - 180 = 20 Tastenzählerwerte Toleranz beim Halten des Tasters

	; Zählerstand erhöhen
	rcall	INC_COUNTER

	; Idle-Zeit resetten
	clr		IDLE_QSECONDS
	clr		IDLE_MINUTES


; Display-Multiplexing
MUX:
	; Evtl. führende Null (Zehnerstelle) verhindern
	cpi		COUNTER, 100
	brsh	NO_LEADING_ZERO

	cpi		DIGIT_1, 0
	brne	NO_LEADING_ZERO

	ldi		DIGIT_1, 10	; Zeichen 10 (also nichts) anzeigen


NO_LEADING_ZERO:
	; Aktuell benötigte Anode mit VCC verbinden
	SET_ANODE:
		ldi		r16, (1 << BUTTON)	; Taster-Pull-Up an; DATA = CLOCK = STROBE = L
		ldi		r17, (1 << BUTTON)

		cpi		MUX_DIGIT, 0
		brne	ACTIVE_DIGIT_1_2_ANODE

	ACTIVE_DIGIT_0_ANODE:
		ori		r16, (1 << DATA)	; Anode für Einerstelle = VCC
		rjmp	SET_ANODE_SHIFTREG

	ACTIVE_DIGIT_1_2_ANODE:
		ori		r17, (1 << DATA)	; Anode für Zehner- und Hunderterstelle = VCC

	SET_ANODE_SHIFTREG:
		out		PORTB, r16
		ori		r16, (1 << CLOCK)	; Clock H
		out		PORTB, r16
		andi	r16, ~(1 << CLOCK)	; Clock L
		out		PORTB, r16

		out		PORTB, r17
		ori		r17, (1 << CLOCK)	; Clock H
		out		PORTB, r17
		andi	r17, ~(1 << CLOCK)	; Clock L
		out		PORTB, r17


	; Hier wird bei Bedarf die Kathode des aktuell zu multiplexenden
	; Segments der Hunderterstelle mit GND verbunden
	; (Die Hunderterstelle hat ihre eigenen beiden Kathoden, braucht also eine Extrawurst)
	ldi		r17, (1 << (1))

	SET_CATHODE_HUNDRED:
		ldi		r16, (1 << BUTTON) | (1 << DATA)	; Taster-Pull-Up an; DATA = H; CLOCK = STROBE = L

		; Sollte die Hunderterstelle gerade inaktiv sein, so muss hier nichts weiter getan werden
		cpi		MUX_DIGIT, 2
		brne	NOTHING_TO_DO_HUNDRED

		; Auch wenn das aktuell in diesem Schleifendurchlauf zu behandelnde Segment nicht mit
		; dem aktuell zu multiplexenden Segment übereinstimmt, muss hier nichts weiter getan werden
		cp		r17, MUX_SEGMENT
		brne	NOTHING_TO_DO_HUNDRED

		; Dazu muss allerdings ersteinmal die zum Inhalt von DIGIT_2 zugeordnete Bitmaske geladen werden
		clr		r19
		cpi		DIGIT_2, 0
		breq	DIGIT_2_BITMASK_DO_NOTHING
		ldi		r19, 0b00000011

	DIGIT_2_BITMASK_DO_NOTHING:
		; Ansonsten prüfen, ob aktuell zu multiplexendes Segment aktiviert werden soll
		mov		r18, MUX_SEGMENT
		and		r18, r19
		tst		r18
		breq	NOTHING_TO_DO_HUNDRED

		; Wenn ja, entsprechende Kathode mit GND verbinden
		andi	r16, ~(1 << DATA)

	NOTHING_TO_DO_HUNDRED:
		out		PORTB, r16
		ori		r16, (1 << CLOCK)	; Clock H
		out		PORTB, r16
		andi	r16, ~(1 << CLOCK)	; Clock L
		out		PORTB, r16

		; r17 weiter nach Rechts shiften; wenn das Bit herausfällt, ist die Schleife beendet
		lsr		r17
		brne	SET_CATHODE_HUNDRED


	; Hier wird bei Bedarf die Kathode des aktuell zu multiplexenden
	; Segments aus einer der ersten beiden Ziffern mit GND verbunden
	ldi		r17, (1 << (6))

	SET_CATHODE_OTHER:
		ldi		r16, (1 << BUTTON) | (1 << DATA)	; Taster Pull-Up an; DATA = H; CLOCK = STROBE = L

		; Sollte gerade die Hunderterstelle aktiv sein, so muss hier nichts weiter getan werden
		cpi		MUX_DIGIT, 2
		breq	NOTHING_TO_DO_OTHER

		; Auch wenn das aktuell in diesem Schleifendurchlauf zu behandelnde Segment nicht mit
		; dem aktuell zu multiplexenden Segment übereinstimmt, muss hier nichts weiter getan werden
		cp		r17, MUX_SEGMENT
		brne	NOTHING_TO_DO_OTHER

		; Ansonsten prüfen, ob das aktuell zu multiplexendes Segment aktiviert werden soll

		; Davor muss allerdings zuerst die richtige Bitmaske zum
		; Inhalt von DIGIT_0 bzw. DIGIT_1 ausgewählt werden
		ldi		zl, low(DIGIT_MASK * 2)
		ldi		zh, high(DIGIT_MASK * 2)

		cpi		MUX_DIGIT, 0
		breq	MUX_DIGIT_0_SELECT

		; Entweder DIGIT_1...
		mov		r19, DIGIT_1
		rjmp	MUX_DIGIT_0_SELECT_END

		; ...oder DIGIT_0 als Bitmaske auswählen
		MUX_DIGIT_0_SELECT:
			mov		r19, DIGIT_0
		MUX_DIGIT_0_SELECT_END:

		; Bitmaske aus dem ROM laden
		lsl		r19, 1
		add		zl, r19
		lpm		r19, z

		; Testen, ob das aktuell zu multiplexende Segment aktiv ist
		and		r19, MUX_SEGMENT
		tst		r19
		breq	NOTHING_TO_DO_OTHER

		; Wenn das aktuelle Segment aktiviert werden soll, entsprechende Kathode mit GND verbinden
		andi	r16, ~(1 << DATA)

	NOTHING_TO_DO_OTHER:
		out		PORTB, r16
		ori		r16, (1 << CLOCK)	; Clock H
		out		PORTB, r16
		andi	r16, ~(1 << CLOCK)	; Clock L
		out		PORTB, r16

		; r17 weiter nach Rechts shiften; wenn das Bit herausfällt, ist die Schleife beendet
		lsr		r17
		brne	SET_CATHODE_OTHER


	; Inhalt der Schieberegister freigeben
	ldi		r16, (1 << BUTTON)	; Taster Pull-Up an; DATA = CLOCK = STROBE = L
	ori		r16, (1 << STROBE)	; Strobe H
	out		PORTB, r16
	andi	r16, ~(1 << STROBE)	; Strobe L
	out		PORTB, r16


	; Auf zum nächsten Segment!
	lsl		MUX_SEGMENT

	; Falls MUX_DIGIT = 2 und Segment > 1, dann Segment = 0 und nächste Ziffer
	cpi		MUX_DIGIT, 2
	brne	MUX_DIGIT_0_1_RESET

	sbrs	MUX_SEGMENT, 2
	rjmp	RETURN

	ldi		MUX_SEGMENT, (1 << (0))
	rjmp	INC_MUX_DIGIT

	; Falls MUX_DIGIT != 2 und Segment > 6, dann Segment = 0 und nächste Ziffer
MUX_DIGIT_0_1_RESET:
	sbrs	MUX_SEGMENT, 7
	rjmp	RETURN

	ldi		MUX_SEGMENT, (1 << (0))

; Nächste zu multiplexende Ziffer auswählen
INC_MUX_DIGIT:
	inc		MUX_DIGIT

	; Falls MUX_DIGIT >= 3, MUX_DIGIT = 0
	cpi		MUX_DIGIT, 3
	breq	MUX_DIGIT_RESET
	rjmp	RETURN

; Hier ist ein kompletter Multiplexingvorgang abgeschlossen
MUX_DIGIT_RESET:
	ldi		MUX_DIGIT, 0

	; Prüfen, ob bereits genug Idle-Zeit verstrichen ist, um das Gerät abzuschalten
	ldi		r16, IDLE_AUTOSHUTDOWN
	cp		IDLE_MINUTES, r16
	brlo	RETURN	; Wenn nicht, weitermachen

	; Schieberegister mit Nullbits füllen (Display ausschalten)
	rcall	SHIFTREG_CLEAR

	; Abschalten (Schlafmodus)
	rcall	SHUTDOWN

	; Und hier geht es nach dem Aufwachen dann auch schon gleich wieder weiter

	; Warten, bis der Taster wieder losgelassen wurde und Taster entprellen
	rcall	BUTTON_DEBOUNCE_WAIT

	; Idle-Zeit resetten
	clr		IDLE_QSECONDS
	clr		IDLE_MINUTES

RETURN:
	rjmp	MAIN


; Zählerstand erhöhen
; (Achtung: Der Zählerstand darf NUR mit INC_COUNTER oder DEC_COUNTER modifiziert werden)
INC_COUNTER:
	inc		COUNTER

	; Einerstelle erhöhen
	inc		DIGIT_0
	cpi		DIGIT_0, 10
	brlo	INC_COUNTER_DONE

	; Übertrag, Zehnerstelle erhöhen
	ldi		DIGIT_0, 0
	inc		DIGIT_1
	cpi		DIGIT_1, 10
	brlo	INC_COUNTER_DONE

	; Übertrag, Hunderterstelle erhöhen
	ldi		DIGIT_1, 0
	inc		DIGIT_2
	cpi		DIGIT_2, 2
	brlo	INC_COUNTER_DONE

	; Wenn Counter > 199, dann Counter = 0
	ldi		COUNTER, 0
	ldi		DIGIT_0, 0
	ldi		DIGIT_1, 0
	ldi		DIGIT_2, 0

INC_COUNTER_DONE:
	ret


; Zählerstand erniedgrigen
; (Achtung: Der Zählerstand darf NUR mit INC_COUNTER oder DEC_COUNTER modifiziert werden)
DEC_COUNTER:
	dec		COUNTER

	; Einerstelle erniedrigen
	dec		DIGIT_0
	cpi		DIGIT_0, 255
	brlo	DEC_COUNTER_DONE

	; Untertrag, Zehnerstelle erniedrigen
	ldi		DIGIT_0, 9
	dec		DIGIT_1
	cpi		DIGIT_1, 255
	brlo	DEC_COUNTER_DONE

	; Untertrag, Hunderterstelle erniedrigen
	ldi		DIGIT_1, 9
	dec		DIGIT_2
	cpi		DIGIT_2, 255
	brlo	DEC_COUNTER_DONE

	; Wenn Counter < 0, dann Counter = 199
	ldi		COUNTER, 199
	ldi		DIGIT_0, 9
	ldi		DIGIT_1, 9
	ldi		DIGIT_2, 1

DEC_COUNTER_DONE:
	ret


; Schieberegister mit Nullbits füllen (Display ausschalten)
SHIFTREG_CLEAR:
	; 11 Nullbits hineinschieben
	ldi		r17, 11

	SHIFTREG_CLEAR_LOOP:
		ldi		r16, (1 << BUTTON)	; Taster-Pull-Up an; DATA = CLOCK = STROBE = L

		out		PORTB, r16
		ori		r16, (1 << CLOCK)	; Clock H
		out		PORTB, r16
		andi	r16, ~(1 << CLOCK)	; Clock L
		out		PORTB, r16

		dec		r17
		brne	SHIFTREG_CLEAR_LOOP

	; Inhalt der Schieberegister freigeben
	ldi		r16, (1 << BUTTON)	; Taster Pull-Up an; DATA = CLOCK = STROBE = L
	ori		r16, (1 << STROBE)	; Strobe H
	out		PORTB, r16
	andi	r16, ~(1 << STROBE)	; Strobe L
	out		PORTB, r16

	ret


; Warten, bis der Taster wieder losgelassen wurde und Taster entprellen
BUTTON_DEBOUNCE_WAIT:
	; Zwei Durchläufe: Entprellen per Delay und auf Loslassen warten
	ldi		r16, 2

	BUTTON_DEBOUNCE_WAIT_LOOP:
		; Delay 1 (wiederholt das kurze Delay)
		ldi		r17, 200

		BUTTON_DEBOUNCE_WAIT_1:
			; Delay 2 (kurzes Delay)
			ldi		r18, 255

			BUTTON_DEBOUNCE_WAIT_2:
				dec		r18
				brne	BUTTON_DEBOUNCE_WAIT_2

			dec		r17
			brne	BUTTON_DEBOUNCE_WAIT_1

		BUTTON_LET_GO:	; Warten, bis Taste losgelassen wurde
			sbis	PINB, BUTTON
			rjmp	BUTTON_LET_GO

	dec		r16
	brne	BUTTON_DEBOUNCE_WAIT_LOOP

	reti


; Abschalten (Schlafmodus)
SHUTDOWN:
	; Ab in den Schlafmodus
	ldi		r16, (1 << SM1) | (1 << SE)	; Power-down Mode
	out		MCUCR, r16

	sleep	; Gute Nacht...

	; Und hier geht es nach dem Aufwachen dann auch schon gleich wieder weiter
	reti


; Siebensegment-Zeichensatz
; Dieser ist aus Faulheit auf der letzten ROM-Page platziert,
; damit ich beim Holen des Zeichens nur ZL verändern muss
.org ((1024 - 22) / 2)

DIGIT_MASK:
	.db 0b0111111, 0	; 0
	.db 0b0000110, 0	; 1
	.db 0b1011011, 0	; 2
	.db 0b1001111, 0	; 3
	.db 0b1100110, 0	; 4
	.db 0b1101101, 0	; 5
	.db 0b1111101, 0	; 6
	.db 0b0000111, 0	; 7
	.db 0b1111111, 0	; 8
	.db 0b1101111, 0	; 9
	.db	0b0000000, 0	; 10 (Leer)
